#ifndef CONFIG_H
#define CONFIG_H

#define APP_NAME        "GPXSee"
#define APP_HOMEPAGE    "http://www.gpxsee.org"
#define DEM_TILES_URL   "http://dem.gpxsee.org/$lat/$lat$lon.hgt.zip"

#endif /* CONFIG_H */
