#ifndef ENC_OBJECTS_H
#define ENC_OBJECTS_H

#define TYPE(t) ((t)<<16)
#define SUBTYPE(t, s) (((t)<<16)|(s))

#define AIRARE 2
#define ACHBRT 3
#define ACHARE 4
#define BCNCAR 5
#define BCNISD 6
#define BCNLAT 7
#define BCNSAW 8
#define BCNSPP 9
#define BERTHS 10
#define BRIDGE 11
#define BUISGL 12
#define BUAARE 13
#define BOYCAR 14
#define BOYINB 15
#define BOYISD 16
#define BOYLAT 17
#define BOYSAW 18
#define BOYSPP 19
#define CBLARE 20
#define CBLOHD 21
#define CBLSUB 22
#define CANALS 23
#define CGUSTA 29
#define COALNE 30
#define CONZNE 31
#define CONVYR 34
#define CRANES 35
#define CURENT 36
#define DAMCON 38
#define DWRTPT 41
#define DEPARE 42
#define DEPCNT 43
#define DRGARE 46
#define DRYDOC 47
#define DMPGRD 48
#define DYKCON 49
#define FAIRWY 51
#define FNCLNE 52
#define FERYRT 53
#define FLODOC 57
#define FOGSIG 58
#define GATCON 61
#define HRBFAC 64
#define HULKES 65
#define LAKARE 69
#define LNDARE 71
#define LNDELV 72
#define LNDRGN 73
#define LNDMRK 74
#define LIGHTS 75
#define LOKBSN 79
#define MARKUL 82
#define MORFAC 84
#define NAVLNE 85
#define OBSTRN 86
#define OFSPLF 87
#define PILPNT 90
#define PILBOP 91
#define PIPARE 92
#define PIPOHD 93
#define PIPSOL 94
#define PONTON 95
#define PRCARE 96
#define PYLONS 98
#define RADRFL 101
#define RADSTA 102
#define RTPBCN 103
#define RDOCAL 104
#define RDOSTA 105
#define RAILWY 106
#define RCRTCL 108
#define RECTRC 109
#define RCTLPT 110
#define RSCSTA 111
#define RESARE 112
#define RIVERS 114
#define ROADWY 116
#define SEAARE 119
#define SLCONS 122
#define SISTAT 123
#define SILTNK 125
#define SLOTOP 126
#define SMCFAC 128
#define SOUNDG 129
#define TSELNE 145
#define TSSBND 146
#define TSSLPT 148
#define TSEZNE 150
#define TUNNEL 151
#define UWTROC 153
#define UNSARE 154
#define WATTUR 156
#define WEDKLP 158
#define WRECKS 159
#define M_COVR 302

#define I_ACHBRT 17000
#define I_ACHARE 17001
#define I_DISMAR 17004
#define I_RESARE 17005
#define I_SISTAT 17007
#define I_BERTHS 17010
#define I_BRIDGE 17011
#define I_CBLOHD 17012
#define I_FERYRT 17013
#define I_HRBFAC 17015
#define I_LOKBSN 17016
#define I_RDOCAL 17017
#define I_HULKES 17020
#define I_PONTON 17021
#define I_PIPOHD 17024
#define I_FLODOC 17025
#define I_BCNLAT 17028
#define I_BOYLAT 17029
#define I_CRANES 17030
#define I_GATCON 17031
#define I_SLCONS 17032
#define I_WTWAXS 17051
#define I_TERMNL 17064
#define I_TRNBSN 17065
#define I_WTWGAG 17067

#endif // ENC_OBJECTS_H
